#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <err.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "dvview.h"
#include "ctrl/usb_ctrl.h"

static const char rcsid[] = "$Id: dvview.c,v 1.3 2004/07/25 22:01:17 a-gota Exp $";

#define	max(a, b)	(a < b ? b : a)

#define	PLAYDV		"/usr/local/bin/playdv"
#define	DIF_BLK_SIZE		    80
#define	DIF_NUM_BLKS_IN_SEQ	   150
#define	DIF_SEQ_SIZE		(DIF_BLK_SIZE * DIF_NUM_BLKS_IN_SEQ)
#define	NUM_DIF_SEQ_525		    10
#define	NUM_DIF_SEQ_625		    12
#define	MAX_FRAME_SIZE	(DIF_SEQ_SIZE * max(NUM_DIF_SEQ_525, NUM_DIF_SEQ_625))

#define	HID_DEVNAME	"/dev/uhid0"


int
main(int argc, char *argv[])
{
    char filenamebuf[FILENAME_MAX], *filename = NULL, *ptr;
    u_char *frame;
    FILE *fp = NULL, *fp_stdin;
    int controller_fd, num_fd;
    int command, play_f, eof_f, exit_f = 0, pipefd[2], verbose_f;
    int param, skip, div, wait_frame;
    long frame_size = 0, num_frame = 0, frame_no = 0;
    int system_is_625 = 0;
    pid_t pid;
    fd_set readfds;
    struct timeval *timeout = NULL, timeout_0, timeout_50, timeout_60;
    struct termios t_orig, t_raw;
    struct stat stat;

    int get_key_event(FILE *fp);
    void parse_subcode(u_char *frame, int system_is_625);

    pipe(pipefd);
    if ((pid = fork()) == 0) {
	/* child */
	dup2(pipefd[0], 0);
	close(pipefd[0]);
	execl(PLAYDV, PLAYDV, NULL);
	return 1;
    }

    if ((frame = malloc(MAX_FRAME_SIZE)) == NULL) {
	kill(pid, SIGTERM);
	waitpid(pid, &command, WNOHANG);
	err(1, "malloc for frame memory");
    }

    if (2 <= argc) {
	filename = argv[1];
	if ((fp = fopen(filename, "r")) == NULL) {
	    warn("%s", filename);
	    filename = NULL;
	} else {
	    fread(frame, DIF_BLK_SIZE, 1, fp);
	    if ((frame[0] & 0xe0) != 0) {
		/* Header is not found */
		fprintf(stderr, "%s: broken DV stream\n", filename);
		fclose(fp);
		fp = NULL;
		filename = NULL;
	    } else {
		fseek(fp, 0L, SEEK_SET);
	    }
	}
    }

    timeout_0.tv_sec = 0;
    timeout_0.tv_usec = 0;
    timeout_50.tv_sec = 0;
    timeout_50.tv_usec = 20000;
    timeout_60.tv_sec = 0;
    timeout_60.tv_usec = 16666;

    if (fp != NULL) {
	printf("DV stream file %s is opened; ", filename);
	if ((system_is_625 = (frame[3] & 0x80)) == 0) {
	    printf("525/60 system");
	    timeout = &timeout_60;
	    frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_525);
	} else {
	    printf("625/50 system");
	    timeout = &timeout_50;
	    frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_625);
	}
	fstat(fileno(fp), &stat);
	num_frame = stat.st_size / frame_size;
	frame_no = 0L;
	printf(", %ld frame(s)\n", num_frame);
    } else {
	printf("No DV stream file is opened\n");
    }

    play_f = 0;
    skip = 0;
    div = 1;
    wait_frame = 0;
    eof_f = 0;
    exit_f = 0;
    verbose_f = 1;

    fp_stdin = fopen("/dev/tty", "r");
    tcgetattr(fileno(fp_stdin), &t_orig);
    t_raw = t_orig;
    t_raw.c_lflag &= ~(ECHO|ICANON);
    tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_raw);

    controller_fd = controller_init(HID_DEVNAME);
    num_fd = max(controller_fd, fileno(fp_stdin)) + 1;

    do {
	FD_ZERO(&readfds);
	FD_SET(fileno(fp_stdin), &readfds);
	if (0 < controller_fd) {
	    FD_SET(controller_fd, &readfds);
	}
	select(num_fd, &readfds, NULL, NULL,
	       play_f ? (1 < div ? timeout : &timeout_0) : NULL);
	command = ACT_NONE;
	if (FD_ISSET(fileno(fp_stdin), &readfds)) {
	    command = get_key_event(fp_stdin);
	}
	if (command == ACT_NONE && 0 < controller_fd && FD_ISSET(controller_fd, &readfds)) {
	    command = controller_get_event(controller_fd);
	}
	if (command != ACT_NONE) {
	    param = (char)(command & ACT_PARAM);

	    if (fp != NULL) {
		switch (command & ACT_MASK) {
		  case ACT_PAUSE:
		    if (play_f) {
			play_f = -1;
		    }
		    div = 1;
		    skip = 0;
		    printf("\nPause\n");
		    break;
		  case ACT_PLAY:
		    if (param == 0) {
			if (play_f) {
			    if (div == 1 && skip == 0) {
				play_f = 0;
				printf("\nPause\n");
				break;
			    }
			} else {
			    if (eof_f) {
				fseek(fp, 0L, SEEK_SET);
				eof_f = 0;
				frame_no = 0L;
			    }
			}
			param = 1;
		    }
		    printf("\nPlay ");
		    play_f = 1;
		    if (0 < param) {
			skip = param - 1;
			div = 1;
			printf("X%d\n", param);
		    } else {
			skip = 0;
			div = -param;
			printf("1/%d\n", -param);
		    }
		    break;
		  case ACT_BACK:
		    printf("\nBackward Play ");
		    play_f = 1;
		    if (0 < param) {
			skip = -(param + 1);
			div = 1;
			printf("X%d\n", param);
		    } else {
			skip = -2;
			div = -param;
			printf("1/%d\n", -param);
		    }
		    break;
		  case ACT_STOP:
		    printf("\nStop\n");
		    play_f = -1;
		    skip = 0;
		    fseek(fp, 0L, SEEK_SET);
		    break;
		  case ACT_NEXT:
		    printf("\nNext frame\n");
		    play_f = -1;
		    skip = 0;
		    break;
		  case ACT_PREV:
		    printf("\nPrevious frame\n");
		    play_f = -1;
		    skip = 0;
		    if (eof_f) {
			fseek(fp, frame_size * (num_frame - 1), SEEK_SET);
			eof_f = 0;
		    } else {
			if ((frame_no -= 2) < 0) {
			    frame_no = 0L;
			    fseek(fp, 0L, SEEK_SET);
			} else {
			    fseek(fp, frame_size * -2, SEEK_CUR);
			}
		    }
		    break;
		  case ACT_REFRESH:
		    printf("\nRefresh last frame\n");
		    play_f = -1;
		    skip = 0;
		    if (eof_f) {
			frame_no = num_frame - 1;
			fseek(fp, frame_size * frame_no, SEEK_SET);
			eof_f = 0;
		    } else {
			if (--frame_no < 0) {
			    frame_no = 0L;
			    fseek(fp, 0L, SEEK_SET);
			} else {
			    fseek(fp, frame_size * -1, SEEK_CUR);
			}
		    }
		    break;
		}
	    }
	    switch (command & ACT_MASK) {
	      case ACT_OPEN:
		if (fp != NULL) {
		    fclose(fp);
		    fp = NULL;
		}
		tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_orig);
		printf("\nOpen file\nDV stream file: ");
		fflush(stdout);
		fgets(filenamebuf, FILENAME_MAX, fp_stdin);
		tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_raw);
		if ((ptr = strchr(filenamebuf, '\n')) != NULL) {
		    *ptr = '\0';
		}
		if ((fp = fopen(filenamebuf, "r")) == NULL) {
		    warn("%s", filenamebuf);
		    filename = NULL;
		} else {
		    filename = filenamebuf;
		    fread(frame, DIF_BLK_SIZE, 1, fp);
		    if ((frame[0] & 0xe0) != 0) {
			/* Header is not found */
			fprintf(stderr, "%s: broken DV stream\n", filename);
			fclose(fp);
			fp = NULL;
			filename = NULL;
		    } else {
			printf("%s is opened; ", filename);
			fseek(fp, 0L, SEEK_SET);
			if ((system_is_625 = (frame[3] & 0x80)) == 0) {
			    printf("525/60 system");
			    timeout = &timeout_60;
			    frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_525);
			} else {
			    printf("625/50 system");
			    timeout = &timeout_50;
			    frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_625);
			}
			fstat(fileno(fp), &stat);
			num_frame = stat.st_size / frame_size;
			printf(", %ld frame(s)\n", num_frame);
			frame_no = 0L;
		    }
		}
		play_f = 0;
		skip = 0;
		break;
	      case ACT_DISP:
		if ((verbose_f = !verbose_f)) {
		    printf("\nVerbose: On\n");
		} else {
		    printf("\nVerbose: Off\n");
		}
		break;
	      case ACT_QUIT:
		play_f = 0;
		exit_f = 1;
		break;
	    }
	}

	if (play_f && fp == NULL) {
	    fprintf(stderr, "No DV stream file is assumed\n");
	}

	if (play_f) {
	    if (wait_frame++ < div) {
		continue;
	    }
	    wait_frame = 0;
	    if (fread(frame, frame_size, 1, fp) < 1) {
		++eof_f;
		play_f = -1;
	    } else {
		printf("#%05ld", frame_no);
		if (verbose_f) {
		    parse_subcode(frame, system_is_625);
		}
		putchar('\r');
		fflush(stdout);
/*
		if (write(pipefd[1], frame, frame_size) < frame_size) {
		    ++exit_f;
		}
*/
		if ((frame_no += (skip + 1)) < 0) {
		    frame_no = 0L;
		    fseek(fp, 0L, SEEK_SET);
		    play_f = 0;
		    printf("\nPause\n");
		} else if (skip != 0) {
		    fseek(fp, frame_size * skip, SEEK_CUR);
		}
	    }
	    if (play_f < 0) {
		++play_f;
	    }
	}
    } while (!exit_f);

    if (fp != NULL) {
	fclose(fp);
    }
    printf("\nClean up...\n");
    kill(pid, SIGTERM);
    waitpid(pid, &command, WNOHANG);

    return 0;
}

struct key_table {
    int chr;
    int action;
};

struct key_table key_table[] = {
    { 'p', ACT_PLAY },
    { '8', ACT_PLAYx8 },
    { '4', ACT_PLAYx4 },
    { '2', ACT_PLAYx2 },
    { '1', ACT_PLAYx1 },
    { '9', ACT_PLAYs5 },
    { '0', ACT_PLAYs10 },
    { ')', ACT_BACKs10 },
    { '(', ACT_BACKs5 },
    { '!', ACT_BACKx1 },
    { '@', ACT_BACKx2 },
    { '$', ACT_BACKx4 },
    { '*', ACT_BACKx8 },
    { 's', ACT_STOP },
    { 'n', ACT_NEXT },
    { 'b', ACT_PREV },
    { 'r', ACT_REFRESH },
    { 'o', ACT_OPEN },
    { 'A', ACT_OPEN_1 },
    { 'B', ACT_OPEN_2 },
    { 'C', ACT_OPEN_3 },
    { 'D', ACT_OPEN_4 },
    { 'E', ACT_OPEN_5 },
    { 'F', ACT_OPEN_6 },
    { 'G', ACT_OPEN_7 },
    { 'H', ACT_OPEN_8 },
    { 'v', ACT_DISP },
    { 'q', ACT_QUIT },
    { 'x', ACT_QUIT },
};


int
get_key_event(FILE *fi)
{
    int c;
    struct key_table *ptr;

    c = fgetc(fi);
    for (ptr = key_table; 0 < ptr->chr; ptr++) {
	if (c == ptr->chr) {
	    return ptr->action;
	}
    }
    return ACT_NONE;
}


void
parse_subcode(u_char *frame, int system_is_625)
{
    int seq, blk, cnt, tmp;
    int num_seq = system_is_625 ? NUM_DIF_SEQ_625 : NUM_DIF_SEQ_525;
    int timecode_f = 0, datacode_f = 0, src_f = 0, date_f = 0, time_f = 0;
    u_char *DIF = frame, *ptr, *PC;
    char *source = NULL, buf[10], date[25], time[15];

    for (seq = 0; seq < num_seq; seq++) {
	for (blk = 0; blk < DIF_NUM_BLKS_IN_SEQ; blk++) {
	    switch ((DIF[0] >> 5) & 0x7) {
	      case 0:	/* Header */
		break;
	      case 1:	/* Subcode */
		if (!timecode_f) {
		    for (ptr = &DIF[3], cnt = 0; cnt < 6; cnt++, ptr += 8) {
			PC = &ptr[3];
			if (PC[0] == 0x13) {
			    /* Timecode pack */
			    printf(": %d%x:%d%x:%d%x.%d%x",
				   (PC[4] >> 4) & 0x03,
				    PC[4] & 0x0f,
				   (PC[3] >> 4) & 0x07,
				    PC[3] & 0x0f,
				   (PC[2] >> 4) & 0x07,
				    PC[2] & 0x0f,
				   (PC[1] >> 4) & 0x03,
				    PC[1] & 0x0f);
			    if (!system_is_625) {
				printf(" %s", (PC[1] & 0x40) ? "DF" : "NDF");
			    }
			    ++timecode_f;
			    break;
			}
		    }
		}
		break;
	      case 2:	/* VAUX */
		if (!datacode_f) {
		    for (ptr = &DIF[3], cnt = 0; cnt < 15; cnt++, ptr += 5) {
			PC = ptr;
			if (PC[0] == 0x60) {
			    /* VAUX Source pack */
			    if (!src_f) {
				tmp = PC[2] & 0x07;
				if (PC[1] == 0xff && tmp == 0x07) {
				    source = "---";
				} else {
				    tmp = tmp * 100
					+ (PC[1] >> 4) * 10
					+ (PC[1] & 0x0f);
				    switch (PC[3] >> 6) {
				      case 0:	/* Camera */
					source = "Camera";
					break;
				      case 1:	/* Line */
					sprintf(buf, "LINE %d", tmp);
					source = buf;
					break;
				      case 2:	/* CATV */
					sprintf(buf, "CH C%d", tmp);
					source = buf;
					break;
				      case 3:	/* TV */
					sprintf(buf, "CH %d", tmp);
					source = buf;
					break;
				    }
				}
				++src_f;
			    }
			} else if (PC[0] == 0x62) {
			    /* VAUX Rec date pack */
			    if (!date_f) {
				sprintf(date, "    %x%x/%02x/%d%x",
				       (PC[4] >> 4) & 0x0f,
					PC[4] & 0x0f,
					PC[3] & 0x0f,
				       (PC[2] >> 4) & 0x03,
					PC[2] & 0x0f);
				++date_f;
			    }
			} else if (PC[0] == 0x63) {
			    /* VAUX Rec time pack */
			    if (!time_f) {
				if ((PC[1] & 0x3f) == 0x3f) {
				    tmp = 0;
				} else {
				    tmp = ((PC[1] >> 4) & 0x03) * 10
					+ (PC[1] & 0x0f);
				}
				sprintf(time, " %d%x:%d%x:%d%x.%02d  ",
					(PC[4] >> 4) & 0x03,
					 PC[4] & 0x0f,
					(PC[3] >> 4) & 0x07,
					 PC[3] & 0x0f,
					(PC[2] >> 4) & 0x07,
					 PC[2] & 0x0f,
					 tmp);
				++time_f;
				break;
			    }
			}
			if (src_f && date_f && time_f) {
			    fputs(date, stdout);
			    fputs(time, stdout);
			    fputs(source, stdout);
			    ++datacode_f;
			    break;
			}
		    }
		}
		break;
	      case 3:	/* Audio */
		break;
	      case 4:	/* Video */
		break;
	    }
	    DIF += DIF_BLK_SIZE;
	}
    }
}
